# This file *must* have the name GSPyModule.py.
#
# The dynamically-linked library (DLL) GSPy.pyd forms the interface between
# GoldSim and this custom Python module. GSPy.pyd imports GSPyModule.py and
# calls the following functions (which *must* be defined in this file):
#
#   - ReturnPyModuleVersion()
#   - NumberOfInputs()
#   - NumberOfOutputs()
#   - DoCalcsAndReturnValues(*inputsList)
#
# The following functions are optional:
#
#   - InitializeSimulation()
#   - FinalizeSimulation()
#
# GSPy.pyd also has a Python API that allows functions (defined in GSPy.pyd)
# to be called from this custom Python module. Include the following command
# to import the Python API:
#
# import GSPy
#
# To relay error messages to GoldSim, use the following function:
#
# GSPy.setErrorMessage(<error message as a string>)
#
# Call the above function within any of the methods defined below.
# For this error message to be displayed by GoldSim, you must return
# a value of 'None'.

# This file developed by Ted Eary (Dec-Jan 2020-2021)
# Import necessary modules
# import <module> as <module abbreviation>
import phreeqpython

# ---------- DEFINE CONSTANTS ---------- #######

# Python module version
PYMODULEVERSION = 0.10

# ******************** USER INPUTS ************************************
# The following constants provide the number of inputs and outputs
# Equal to number of species coming from GoldSim
NUMINPUTS = 21
NUMOUTPUTS = 25

# ----- ESSENTIAL FUNCTIONS TO DEFINE ----- #######

# The following functions are called by GSPy.pyd and therefore *must* be
# defined in GSPyModule.py:
#
#   - ReturnPyModuleVersion(): Returns the version of the custom Python module
#   - NumberOfInputs(): Returns number of inputs needed for custom Python
#     module calculations
#   - NumberOfOutputs(): Returns number of outputs returned by custom Python
#     module calculations
#   - DoCalcsAndReturnValues(*inputsList): Executes the custom calculations
#     specified by the user and returns results
#
# NOTE: The asterisk in *inputsList is essential syntax (but another variable
#       name can be used
#       instead of 'inputsList', if desired). GSPy.pyd passes input arguments
#       as individual scalar
#       values. The asterisk combines these into a tuple.

# Optional initialization function
# def InitializeSimulation():
#   print("Running Python function: InitializeSimulation...")
# <include any initialization commands>
#   return 0


# Function used to query the version of this custom Python module
def ReturnPyModuleVersion():
    return PYMODULEVERSION


# Function used to query the number of inputs needed for the custom Python
#  module calculations
def NumberOfInputs():
    return NUMINPUTS


# Function used to query the number of outputs returned by the custom Python
# module calculations
def NumberOfOutputs():
    return NUMOUTPUTS


# Executes the custom calculations
def DoCalcsAndReturnValues(*inputsList):
    # Get the input concentrations from GoldSim, which is a vector of
    #  species concentrations, and then convert the inputsList tuple to a list
    species_conc = list(inputsList)

    # Get the list of species IDs used in the GoldSim model
    species_ids = get_species_ids()

    # Get the list of gram formula weights of the GoldSim species
    species_gfwts = get_gfwts()

    # Get the GoldSim Species IDs converted to PHREEQC format, applies
    #  mostly to redox species
    converted_species_ids = convert_ids(species_ids)

    # Start building the input to PHREEQC:
    # First, put the solution conditions at the start of the inputs list
    # NOTE: DO NOT remove the space(s) in the key 'pH  6'!  The pH coming
    # from GoldSim in mg/L is purely a result of mixing so has no chemical
    # basis, so it will get removed from inputs_dict using pop. The
    # valid pH from PHREEQC will get returned in the last position of
    # phreeq_results below.  This is one good reason to make sure that
    # the input solution chemistries in GoldSim are ion charged-
    # balanced so that the pH coming back from PHREEQC is a function
    # of specified chemical processes and not charge imbalance.
    inputs_dict = {'units': 'mg/l',
                   'temperature': 25.0,
                   'pH  6': 'charge',
                   'pe': 4.0
                   }

    # Construct the input dictionary to PHREEQC from the two lists
    # of ids and concentrations
    id_conc = dict(zip(converted_species_ids, species_conc))
    inputs_dict.update(id_conc)

    # Remove the entry for key='pH' from list; it will get replaced by result
    # from PHREEQC. This is done to make sure that the pH returned to
    # GoldSim is the result of the PHREEQC calculation.
    inputs_dict.pop('pH')

    # Get an instance of phreeqcpython
    # Databases are located in
    #  ...\Anaconda3 Installation\Lib\site-packages\phreeqpython\database.  The thermodynamic datafile of
    # wateq4f.dat was copied into this folder from phreeqc v.3.0 installation.
    pqpy = phreeqpython.PhreeqPython(database='./wateq4f.dat',
                                     from_file=None, debug=False)

    # Create the solution in PHREEQC -- same idea as using the
    # SOLUTION keyword in PHREEQC
    solution0 = pqpy.add_solution(inputs_dict)

    
    # Equilibrate the solution with solubilities -- same as using the
    # EQUILIBRIUM_PHASES keyword in PHREEQC.
    solution0.equalize(['Al(OH)3(a)', 'Fe(OH)3(a)', 'Gypsum',
                        'O2(g)', 'CO2(g)'],
                       [0, 0, 0, -0.7, -3.5],
                       [0, 0, 0, 1, 1])
    
    # Get results to return to GoldSim, but first remove characters not
    # part of atomic abbreviations so can use phreeqpython methods
    # to get results for the species ids used in GoldSim
    total_ids = [''.join(c for c in i if c not in '()-+0123456')
                 for i in converted_species_ids]

    # Get the results from PHREEQC in units of mole/liter for each element
    # Note, this means for elements like S (sulfur) results are in
    # moles S/liter not moles SO4/liter so need to make the conversion
    # using the atomic weights for each species stoichiometry in the GoldSim
    # species vector if want to get the output from PHREEQC into units
    # of mg/L for return to GoldSim.
    phreeq_results = []
    for element in total_ids:
        phreeq_results.append(solution0.total_element(element, units='mol'))

    # Convert from mole/L to mg/L if needed. User should change logic if
    # other units are needed.
    results_to_GoldSim = [phreeq_results[i] * species_gfwts[i] * 1000
                          for i in range(len(phreeq_results))]

    # Put pH at the end of the results so is not multiplied by 1000 in
    # above unit conversion
    results_to_GoldSim[-1] = solution0.pH

    # Get some saturation indices from PHREEQC and add them to the end
    # of the list of results going back to GoldSim
    SI_values = [solution0.si('Al(OH)3(a)'), solution0.si('Fe(OH)3(a)'),
          solution0.si('Gypsum'), solution0.si('Jarosite(ss')]
    results_to_GoldSim += SI_values

    # Return the result to GSPy.pyd as a tuple
    return tuple(results_to_GoldSim)

# def FinalizeSimulation():
#    print("Running Python function: FinalizeSimulation...")
#    # <include any initialization commands>
#    return 0

# ----- DEFINE ADDITIONAL HELPER FUNCTIONS ----- #######

# The following functions are not called directly by GSPy.pyd, but they are
# essential helper functions to carry out the custom calculations


# <FUNCTIONS DEFINED HERE>

def get_species_ids():
    # ******************** USER INPUT ************************************
    # Create a list with the species IDs used in the GoldSim model. There
    # should be NUMINPUTS number of items in the list and the items need
    # to be in the same order as in the GoldSim species vector.
    ids = ['Al', 'As', 'Ba', 'Ca', 'Cd', 'Cl', 'Cu', 'C_4',
           'Fe', 'K', 'Mg', 'Mn', 'NO3_N', 'Na', 'Ni', 'Pb',
           'SO4', 'Sr', 'Si', 'Zn', 'pH'
           ]
    return ids


def get_gfwts():
    # ******************** USER INPUT ************************************
    # Create a list with the gram formula weights of the GoldSim species
    # as used in phreeqc. There should be NUMINPUTS number of items in
    # the list and the items need to be in the same order as in the
    # GoldSim species vector.
    gfwts = [26.98, 74.92, 137.33, 40.08, 112.41, 35.45, 63.55, 61.0,
             55.85, 39.09, 24.31, 54.94, 14.01, 22.99, 58.69, 207.2,
             96.06, 28.09, 87.62, 65.41, 1.008
             ]
    return gfwts


def convert_ids(ids: list):
    # Sort through the list of common species IDs from GoldSim and convert
    # the ones that do not conform to PHREEQC requirements. These are mostly
    # the redox species

    # gs_ids are typical GoldSim species IDS.  Add more if needed.
    gs_ids = ['C_4', 'CO3', 'CH4', 'SO4', 'H2S', 'NO3_N', 'NO3',
              'NH4_N', 'NH4', 'NO2_N', 'NO2', 'O2', 'N2', 'H2']

    # pq_ids = PHREEQC standard basis species designations for redox elements
    pq_ids = ['C(4)', 'C(4)', 'C(-4)', 'S(6)', 'S(-2)', 'N(5)', 'N(5)',
              'N(-3)', 'N(-3)', 'N(3)', 'N(3)', 'O(0)', 'N(0)', 'H(0)']

    # Go through the list ids and change the ones that match to PHREEQC
    #  type identifiers
    for i, i_item in enumerate(ids):
        for j, j_item in enumerate(gs_ids):
            if i_item == j_item:
                ids[i] = pq_ids[j]

    return ids
