import phreeqpython
# This file meant to simulated sending list of chemical concentrations from
# GoldSim to phreeqpython and return results back to GoldSim.  It can be used
# to experiment with the various phreeqpython functions that connect to
# phreeqc. 
# This file developed by Ted Eary (Dec-Jan 2020-2021)

def convert_ids(ids: list):
    # Sort through the list of common species IDs from GoldSim and convert
    # the ones that do not conform to PHREEQC requirements
    # These are mostly redox-affected species
    # gs_ids = GoldSim species IDS
    gs_ids = ['HCO3', 'CO3', 'CH4', 'SO4', 'H2S', 'NO3_N', 'NO3',
              'NH4_N', 'NH4', 'NO2_N', 'NO2', 'O2', 'N2', 'H2'
              ]
    # pq_ids = PHREEQC standard basis species
    pq_ids = ['C(4)', 'C(4)', 'C(-4)', 'S(6)', 'S(-2)', 'N(5)', 'N(5)',
              'N(-3)', 'N(-3)', 'N(3)', 'N(3)', 'O(0)', 'N(0)', 'H(0)'
              ]

    for i, i_item in enumerate(ids):
        for j, j_item in enumerate(gs_ids):
            if i_item == j_item:
                ids[i] = pq_ids[j]

    print("new ids ", ids)

    return ids


def get_species_ids():
    # Create a list with the species IDs used in the GoldSim model
    ids = ['Al', 'As', 'Ba', 'Ca', 'Cd', 'Cl', 'Cu', 'HCO3',
           'Fe', 'K', 'Mg', 'Mn', 'NO3_N', 'Na', 'Ni', 'Pb',
           'SO4', 'Sr', 'Si', 'Zn', 'pH'
           ]
    return ids


def get_gfwts():
    # Create a list with the gram formula weights of the GoldSim species
    # as used in phreeqc - should be NUMINPUTS number of items in list
    gfwts = [26.98, 74.92, 137.33, 40.08, 112.41, 35.45, 63.55, 61.0,
             55.85, 39.09, 24.31, 54.94, 14.01, 22.99, 58.69, 207.2,
             96.07, 28.09, 87.62, 65.41, 1.008
             ]
    return gfwts


# Get the list of species IDs used in the GoldSim model
species_ids = get_species_ids()

# Get the list of gram formula weights of the GoldSim species
species_gfwts = get_gfwts()

# Get the GoldSim Species IDs converted to PHREEQC format, applies
# mostly to redox species
converted_species_ids = convert_ids(species_ids)

# Concentrations from GoldSim - should be NUMINPUTS number of items in list
species_conc = [45.627, 0.024854, 0.044047, 593.92, 0.0039248,
                14.916, 0.16876, 0.0058719, 0.97124, 55.345,
                17.326, 12.665, 1.5191, 37.896, 0.015715,
                0.082138, 1892.8, 0.0028673, 0.55416, 3.2786,
                4.22267
                ]

# Start building the Solution definition for input to phreeqc. First put
# the units, temperature, and density (optional) at the start of the
# inputs dictionary
inputs_dict = {'units': 'mg/L',
               'temperature': 25.0,
               'pH  6': "charge",
               'pe': 4.0
               }

# Construct the rest of the input dictionary to PHREEQC from two lists
# of converted_ids (from species_ids) and species_conc
id_conc = dict(zip(converted_species_ids, species_conc))
inputs_dict.update(id_conc)
# Remove the entry for key='pH' from list; it will get replaced by result
# from phreeqc
inputs_dict.pop('pH')

# Check inputs to phreeqc
print()
for key, value in inputs_dict.items():
    print("{0} : {1}".format(key, value))

# Get an instance of phreeqcpython
pqpy = phreeqpython.PhreeqPython()

# Databases are located in
# .../Anaconda3 Installation\Lib\site-packages\phreeqpython\database
pqpy = phreeqpython.PhreeqPython(database='./wateq4f.dat',
                                 from_file=None, debug=True)

solution0 = pqpy.add_solution(inputs_dict)
solution0.equalize(['Al(OH)3(a)', 'Fe(OH)3(a)', 'Gypsum',
                    'O2(g)', 'CO2(g)'],
                   [0, 0, 0, -0.7, -3.5],
                   [0, 0, 0, 1, 1])

print(str(solution0.molality('SO4-2', units='mol')))
print(str(solution0.total('S', units='mol')))
print(str(solution0.total_element('S', units='mol')))
print(str(solution0.si('Jarosite(ss)')))

# Check results if needed
for key, value in solution0.elements.items():
    print("{0}: {1}".format(key, value))
    print("pH: {0}".format(solution0.pH))

# Get results to return to GoldSim
# First remove characters not part of atomic abbreviations so can use
# phreeqpython methods to get results for species used in GoldSim
total_ids = [''.join(c for c in i if c not in '()-+0123456')
             for i in converted_species_ids]
print(total_ids)

# Get the results from PHREEQC in units of mole/liter for each element
# Note, this means for elements like S (sulfur) results are in moles S/liter
# not moles SO4/liter so need to make the conversion using the atomic
# weights in species.xlsx if want to use mg/L as the output returned
# to GoldSim.
results = []
for element in total_ids:
    results.append(solution0.total_element(element, units='mol'))
# Put pH at the end of the results
results[-1] = solution0.pH

results_to_GoldSim = [results[i] * species_gfwts[i] * 1000
                      for i in range(len(results))]
# results_to_GoldSim = [results[i] * 1 * 1
#                      for i in range(len(results))]

# Put pH at the end of the results so is not multiplied by 1000
results_to_GoldSim[-1] = solution0.pH

# Check results
print('*'*40)
for key, value in enumerate(results_to_GoldSim):
    print(key, value)

print("Jarosite SI = ", str(solution0.si('Jarosite(ss)')))
